/**
 * data-structure-typed
 *
 * @author Tyler Zeng
 * @copyright Copyright (c) 2022 Tyler Zeng <zrwusa@gmail.com>
 * @license MIT License
 */
import type { AVLTreeMultiMapNested, AVLTreeMultiMapNodeNested, AVLTreeMultiMapOptions, BinaryTreeDeleteResult, BSTNKeyOrNode, BTNCallback, IterationType, KeyOrNodeOrEntry } from '../../types';
import { IBinaryTree } from '../../interfaces';
import { AVLTree, AVLTreeNode } from './avl-tree';
export declare class AVLTreeMultiMapNode<K = any, V = any, NODE extends AVLTreeMultiMapNode<K, V, NODE> = AVLTreeMultiMapNodeNested<K, V>> extends AVLTreeNode<K, V, NODE> {
    /**
     * The constructor function initializes a BinaryTreeNode object with a key, value, and count.
     * @param {K} key - The `key` parameter is of type `K` and represents the unique identifier
     * of the binary tree node.
     * @param {V} [value] - The `value` parameter is an optional parameter of type `V`. It represents the value of the binary
     * tree node. If no value is provided, it will be `undefined`.
     * @param {number} [count=1] - The `count` parameter is a number that represents the number of times a particular value
     * occurs in a binary tree node. It has a default value of 1, which means that if no value is provided for the `count`
     * parameter when creating a new instance of the `BinaryTreeNode` class.
     */
    constructor(key: K, value?: V, count?: number);
    protected _count: number;
    /**
     * The function returns the value of the protected variable _count.
     * @returns The count property of the object, which is of type number.
     */
    get count(): number;
    /**
     * The above function sets the value of the count property.
     * @param {number} value - The value parameter is of type number, which means it can accept any
     * numeric value.
     */
    set count(value: number);
}
/**
 * The only distinction between a AVLTreeMultiMap and a AVLTree lies in the ability of the former to store duplicate nodes through the utilization of counters.
 */
export declare class AVLTreeMultiMap<K = any, V = any, NODE extends AVLTreeMultiMapNode<K, V, NODE> = AVLTreeMultiMapNode<K, V, AVLTreeMultiMapNodeNested<K, V>>, TREE extends AVLTreeMultiMap<K, V, NODE, TREE> = AVLTreeMultiMap<K, V, NODE, AVLTreeMultiMapNested<K, V, NODE>>> extends AVLTree<K, V, NODE, TREE> implements IBinaryTree<K, V, NODE, TREE> {
    constructor(keysOrNodesOrEntries?: Iterable<KeyOrNodeOrEntry<K, V, NODE>>, options?: AVLTreeMultiMapOptions<K>);
    protected _count: number;
    /**
     * The function calculates the sum of the count property of all nodes in a tree using depth-first
     * search.
     * @returns the sum of the count property of all nodes in the tree.
     */
    get count(): number;
    /**
     * Time Complexity: O(n)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(n)
     * Space Complexity: O(1)
     *
     * The function calculates the sum of the count property of all nodes in a tree using depth-first
     * search.
     * @returns the sum of the count property of all nodes in the tree.
     */
    getComputedCount(): number;
    /**
     * The function creates a new BSTNode with the given key, value, and count.
     * @param {K} key - The key parameter is the unique identifier for the binary tree node. It is used to
     * distinguish one node from another in the tree.
     * @param {NODE} value - The `value` parameter represents the value that will be stored in the binary search tree node.
     * @param {number} [count] - The "count" parameter is an optional parameter of type number. It represents the number of
     * occurrences of the value in the binary search tree node. If not provided, the count will default to 1.
     * @returns A new instance of the BSTNode class with the specified key, value, and count (if provided).
     */
    createNode(key: K, value?: V, count?: number): NODE;
    /**
     * The function creates a new AVLTreeMultiMap object with the specified options and returns it.
     * @param [options] - The `options` parameter is an optional object that contains additional
     * configuration options for creating the `AVLTreeMultiMap` object. It can include properties such as
     * `iterationType` and `variant`, which are used to specify the type of iteration and the variant of
     * the tree, respectively. These properties can be
     * @returns a new instance of the `AVLTreeMultiMap` class, with the provided options merged with the
     * default options. The returned value is casted as `TREE`.
     */
    createTree(options?: AVLTreeMultiMapOptions<K>): TREE;
    /**
     * The function `keyValueOrEntryToNode` converts an keyOrNodeOrEntry object into a node object.
     * @param keyOrNodeOrEntry - The `keyOrNodeOrEntry` parameter is of type `KeyOrNodeOrEntry<K, V, NODE>`, which means it
     * can be one of the following:
     * @param {V} [value] - The `value` parameter is an optional argument that represents the value
     * associated with the node. It is of type `V`, which can be any data type. If no value is provided,
     * it defaults to `undefined`.
     * @param [count=1] - The `count` parameter is an optional parameter that specifies the number of
     * times the value should be added to the node. If not provided, it defaults to 1.
     * @returns a node of type `NODE` or `undefined`.
     */
    keyValueOrEntryToNode(keyOrNodeOrEntry: KeyOrNodeOrEntry<K, V, NODE>, value?: V, count?: number): NODE | undefined;
    /**
     * The function checks if an keyOrNodeOrEntry is an instance of the AVLTreeMultiMapNode class.
     * @param keyOrNodeOrEntry - The `keyOrNodeOrEntry` parameter is of type `KeyOrNodeOrEntry<K, V, NODE>`.
     * @returns a boolean value indicating whether the keyOrNodeOrEntry is an instance of the AVLTreeMultiMapNode
     * class.
     */
    isNode(keyOrNodeOrEntry: KeyOrNodeOrEntry<K, V, NODE>): keyOrNodeOrEntry is NODE;
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     *
     * The function overrides the add method of a binary tree node and adds a new node to the tree.
     * @param keyOrNodeOrEntry - The `keyOrNodeOrEntry` parameter can be either a key, a node, or an
     * entry. It represents the key, node, or entry that you want to add to the binary tree.
     * @param {V} [value] - The `value` parameter represents the value associated with the key in the
     * binary tree node. It is an optional parameter, meaning it can be omitted when calling the `add`
     * method.
     * @param [count=1] - The `count` parameter represents the number of times the key-value pair should
     * be added to the binary tree. By default, it is set to 1, meaning that the key-value pair will be
     * added once. However, you can specify a different value for `count` if you want to add
     * @returns The method is returning either the newly inserted node or `undefined` if the insertion
     * was not successful.
     */
    add(keyOrNodeOrEntry: KeyOrNodeOrEntry<K, V, NODE>, value?: V, count?: number): boolean;
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     *
     * The `delete` function in TypeScript is used to remove a node from a binary tree, taking into
     * account the count of the node and balancing the tree if necessary.
     * @param identifier - The identifier is the value or key that is used to identify the node that
     * needs to be deleted from the binary tree. It can be of any type that is returned by the callback
     * function.
     * @param {C} callback - The `callback` parameter is a function that is used to determine if a node
     * should be deleted. It is optional and defaults to a default callback function. The `callback`
     * function takes one parameter, which is the identifier of the node, and returns a value that is
     * used to identify the node to
     * @param [ignoreCount=false] - A boolean flag indicating whether to ignore the count of the node
     * being deleted. If set to true, the count of the node will not be considered and the node will be
     * deleted regardless of its count. If set to false (default), the count of the node will be
     * decremented by 1 and
     * @returns an array of `BinaryTreeDeleteResult<NODE>`.
     */
    delete<C extends BTNCallback<NODE>>(identifier: ReturnType<C>, callback?: C, ignoreCount?: boolean): BinaryTreeDeleteResult<NODE>[];
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The clear() function clears the contents of a data structure and sets the count to zero.
     */
    clear(): void;
    /**
     * Time Complexity: O(n log n)
     * Space Complexity: O(log n)
     */
    /**
     * Time Complexity: O(n log n)
     * Space Complexity: O(log n)
     *
     * The `perfectlyBalance` function takes a sorted array of nodes and builds a balanced binary search
     * tree using either a recursive or iterative approach.
     * @param iterationType - The `iterationType` parameter is an optional parameter that specifies the
     * type of iteration to use when building the balanced binary search tree. It can have two possible
     * values:
     * @returns a boolean value.
     */
    perfectlyBalance(iterationType?: IterationType): boolean;
    /**
     * Time complexity: O(n)
     * Space complexity: O(n)
     */
    /**
     * Time complexity: O(n)
     * Space complexity: O(n)
     *
     * The `clone` function creates a deep copy of a tree object.
     * @returns The `clone()` method is returning a cloned instance of the `TREE` object.
     */
    clone(): TREE;
    /**
     * The `_swapProperties` function swaps the key, value, count, and height properties between two nodes.
     * @param {K | NODE | undefined} srcNode - The `srcNode` parameter represents the source node from
     * which the values will be swapped. It can be of type `K`, `NODE`, or `undefined`.
     * @param {K | NODE | undefined} destNode - The `destNode` parameter represents the destination
     * node where the values from the source node will be swapped to.
     * @returns either the `destNode` object if both `srcNode` and `destNode` are defined, or `undefined`
     * if either `srcNode` or `destNode` is undefined.
     */
    protected _swapProperties(srcNode: BSTNKeyOrNode<K, NODE>, destNode: BSTNKeyOrNode<K, NODE>): NODE | undefined;
    /**
     * The function replaces an old node with a new node and updates the count property of the new node.
     * @param {NODE} oldNode - The `oldNode` parameter is of type `NODE` and represents the node that
     * needs to be replaced in a data structure.
     * @param {NODE} newNode - The `newNode` parameter is an object of type `NODE`.
     * @returns The method is returning the result of calling the `_replaceNode` method from the
     * superclass, after updating the `count` property of the `newNode` object.
     */
    protected _replaceNode(oldNode: NODE, newNode: NODE): NODE;
}
