import type { BinaryTreeDeleteResult, BTNCallback, KeyOrNodeOrEntry, RBTreeOptions, RedBlackTreeNested, RedBlackTreeNodeNested } from '../../types';
import { CP, CRUD, RBTNColor } from '../../types';
import { BST, BSTNode } from './bst';
import { IBinaryTree } from '../../interfaces';
export declare class RedBlackTreeNode<K = any, V = any, NODE extends RedBlackTreeNode<K, V, NODE> = RedBlackTreeNodeNested<K, V>> extends BSTNode<K, V, NODE> {
    /**
     * The constructor function initializes a Red-Black Tree Node with a key, an optional value, and a
     * color.
     * @param {K} key - The key parameter is of type K and represents the key of the node in the
     * Red-Black Tree.
     * @param {V} [value] - The `value` parameter is an optional parameter that represents the value
     * associated with the key in the Red-Black Tree Node. It is not required and can be omitted when
     * creating a new instance of the Red-Black Tree Node.
     * @param {RBTNColor} color - The `color` parameter is used to specify the color of the Red-Black
     * Tree Node. It is an optional parameter with a default value of `'BLACK'`.
     */
    constructor(key: K, value?: V, color?: RBTNColor);
    protected _color: RBTNColor;
    /**
     * The function returns the color value of a variable.
     * @returns The color value stored in the private variable `_color`.
     */
    get color(): RBTNColor;
    /**
     * The function sets the color property to the specified value.
     * @param {RBTNColor} value - The value parameter is of type RBTNColor.
     */
    set color(value: RBTNColor);
}
export declare class RedBlackTree<K = any, V = any, NODE extends RedBlackTreeNode<K, V, NODE> = RedBlackTreeNode<K, V, RedBlackTreeNodeNested<K, V>>, TREE extends RedBlackTree<K, V, NODE, TREE> = RedBlackTree<K, V, NODE, RedBlackTreeNested<K, V, NODE>>> extends BST<K, V, NODE, TREE> implements IBinaryTree<K, V, NODE, TREE> {
    /**
     * This is the constructor function for a Red-Black Tree data structure in TypeScript.
     * @param keysOrNodesOrEntries - The `keysOrNodesOrEntries` parameter is an iterable object that can
     * contain keys, nodes, or entries. It is used to initialize the RBTree with the provided keys,
     * nodes, or entries.
     * @param [options] - The `options` parameter is an optional object that can be passed to the
     * constructor. It allows you to customize the behavior of the RBTree. It can include properties such
     * as `compareKeys`, `compareValues`, `allowDuplicates`, etc. These properties define how the RBTree
     * should compare keys and
     */
    constructor(keysOrNodesOrEntries?: Iterable<KeyOrNodeOrEntry<K, V, NODE>>, options?: RBTreeOptions<K>);
    protected _root: NODE | undefined;
    /**
     * The function returns the root node of a tree or undefined if there is no root.
     * @returns The root node of the tree structure, or undefined if there is no root node.
     */
    get root(): NODE | undefined;
    /**
     * The function creates a new Red-Black Tree node with the specified key, value, and color.
     * @param {K} key - The key parameter represents the key of the node being created. It is of type K,
     * which is a generic type representing the key's data type.
     * @param {V} [value] - The `value` parameter is an optional parameter that represents the value
     * associated with the key in the node. It is not required and can be omitted if not needed.
     * @param {RBTNColor} color - The "color" parameter is used to specify the color of the node in a
     * Red-Black Tree. It is an optional parameter with a default value of "'BLACK'". The color
     * can be either "'RED'" or "'BLACK'".
     * @returns The method is returning a new instance of a RedBlackTreeNode with the specified key,
     * value, and color.
     */
    createNode(key: K, value?: V, color?: RBTNColor): NODE;
    /**
     * The function creates a Red-Black Tree with the given options and returns it.
     * @param [options] - The `options` parameter is an optional object that contains configuration
     * options for creating the Red-Black Tree. It is of type `RBTreeOptions<K>`, where `K` represents
     * the type of keys in the tree.
     * @returns a new instance of a RedBlackTree object.
     */
    createTree(options?: RBTreeOptions<K>): TREE;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function `keyValueOrEntryToNode` takes a key, value, or entry and returns a node if it is
     * valid, otherwise it returns undefined.
     * @param {KeyOrNodeOrEntry<K, V, NODE>} keyOrNodeOrEntry - The key, value, or entry to convert.
     * @param {V} [value] - The value associated with the key (if `keyOrNodeOrEntry` is a key).
     * @returns {NODE | undefined} - The corresponding Red-Black Tree node, or `undefined` if conversion fails.
     */
    keyValueOrEntryToNode(keyOrNodeOrEntry: KeyOrNodeOrEntry<K, V, NODE>, value?: V): NODE | undefined;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     * /
  
     /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function checks if the input is an instance of the RedBlackTreeNode class.
     * @param {KeyOrNodeOrEntry<K, V, NODE>} keyOrNodeOrEntry - The object to check.
     * @returns {boolean} - `true` if the object is a Red-Black Tree node, `false` otherwise.
     */
    isNode(keyOrNodeOrEntry: KeyOrNodeOrEntry<K, V, NODE>): keyOrNodeOrEntry is NODE;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The "clear" function sets the root node of a data structure to a sentinel value and resets the
     * size counter to zero.
     */
    clear(): void;
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     *
     * The function adds a new node to a Red-Black Tree data structure and returns a boolean indicating
     * whether the operation was successful.
     * @param keyOrNodeOrEntry - The `keyOrNodeOrEntry` parameter can be either a key, a node, or an
     * entry.
     * @param {V} [value] - The `value` parameter is the value associated with the key that is being
     * added to the tree.
     * @returns The method is returning a boolean value. It returns true if the node was successfully
     * added or updated, and false otherwise.
     */
    add(keyOrNodeOrEntry: KeyOrNodeOrEntry<K, V, NODE>, value?: V): boolean;
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     *
     * The function `delete` in a binary tree class deletes a node from the tree and fixes the tree if
     * necessary.
     * @param {ReturnType<C> | null | undefined} identifier - The `identifier` parameter is the
     * identifier of the node that needs to be deleted from the binary tree. It can be of any type that
     * is returned by the callback function `C`. It can also be `null` or `undefined` if the node to be
     * deleted is not found.
     * @param {C} callback - The `callback` parameter is a function that is used to retrieve a node from
     * the binary tree based on its identifier. It is an optional parameter and if not provided, the
     * `_DEFAULT_CALLBACK` function is used as the default callback. The callback function should
     * return the identifier of the node to
     * @returns an array of BinaryTreeDeleteResult<NODE> objects.
     */
    delete<C extends BTNCallback<NODE>>(identifier: ReturnType<C> | null | undefined, callback?: C): BinaryTreeDeleteResult<NODE>[];
    /**
     * The function sets the root of a tree-like structure and updates the parent property of the new
     * root.
     * @param {NODE | undefined} v - v is a parameter of type NODE or undefined.
     */
    protected _setRoot(v: NODE | undefined): void;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function replaces an old node with a new node while preserving the color of the old node.
     * @param {NODE} oldNode - The `oldNode` parameter represents the node that needs to be replaced in
     * the data structure.
     * @param {NODE} newNode - The `newNode` parameter is the new node that will replace the old node in
     * the data structure.
     * @returns The method is returning the result of calling the `_replaceNode` method from the
     * superclass, with the `oldNode` and `newNode` parameters.
     */
    protected _replaceNode(oldNode: NODE, newNode: NODE): NODE;
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     *
     * The `_insert` function inserts or updates a node in a binary search tree and performs necessary
     * fix-ups to maintain the red-black tree properties.
     * @param {NODE} node - The `node` parameter represents the node that needs to be inserted into a
     * binary search tree. It contains a `key` property that is used to determine the position of the
     * node in the tree.
     * @returns {'inserted' | 'updated'} - The result of the insertion.
     */
    protected _insert(node: NODE): CRUD;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function `_transplant` is used to replace a node `u` with another node `v` in a binary tree.
     * @param {NODE} u - The parameter "u" represents a node in a binary tree.
     * @param {NODE | undefined} v - The parameter `v` is of type `NODE | undefined`, which means it can
     * either be a `NODE` object or `undefined`.
     */
    protected _transplant(u: NODE, v: NODE | undefined): void;
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     *
     * The `_insertFixup` function is used to fix the Red-Black Tree after inserting a new node.
     * @param {NODE | undefined} z - The parameter `z` represents a node in the Red-Black Tree. It can
     * either be a valid node object or `undefined`.
     */
    protected _insertFixup(z: NODE | undefined): void;
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     *
     * The `_deleteFixup` function is used to fix the red-black tree after a node deletion by adjusting
     * the colors and performing rotations.
     * @param {NODE | undefined} node - The `node` parameter represents a node in a Red-Black Tree data
     * structure. It can be either a valid node object or `undefined`.
     * @returns The function does not return any value. It has a return type of `void`.
     */
    protected _deleteFixup(node: NODE | undefined): void;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The `_leftRotate` function performs a left rotation on a given node in a binary tree.
     * @param {NODE | undefined} x - The parameter `x` is of type `NODE | undefined`. It represents a
     * node in a binary tree or `undefined` if there is no node.
     * @returns void, which means it does not return any value.
     */
    protected _leftRotate(x: NODE | undefined): void;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The `_rightRotate` function performs a right rotation on a given node in a binary tree.
     * @param {NODE | undefined} y - The parameter `y` is of type `NODE | undefined`. It represents a
     * node in a binary tree or `undefined` if there is no node.
     * @returns void, which means it does not return any value.
     */
    protected _rightRotate(y: NODE | undefined): void;
    /**
     * The function compares two values using a comparator function and returns whether the first value
     * is greater than, less than, or equal to the second value.
     * @param {K} a - The parameter "a" is of type K.
     * @param {K} b - The parameter "b" in the above code represents a K.
     * @returns a value of type CP (ComparisonResult). The possible return values are 'GT' (greater
     * than), 'LT' (less than), or 'EQ' (equal).
     */
    protected _compare(a: K, b: K): CP;
}
